"""
This package (along with its subpackages) contains the classes used to model
energy systems. An energy system is modelled as a graph/network of entities
with very specific constraints on which types of entities are allowed to be
connected.

"""
# TODO: Adhere to PEP 0257 by listing the exported classes with a short
#       summary.


class Entity:
    r"""
    The most abstract type of vertex in an energy system graph. Since each
    entity in an energy system has to be uniquely identifiable and
    connected (either via input or via output) to at least one other
    entity, these properties are collected here so that they are shared
    with descendant classes.

    Parameters
    ----------
    uid : string or tuple
        Unique component identifier of the entity.
    inputs : list
        List of Entities acting as input to this Entity.
    outputs : list
        List of Entities acting as output from this Enity.
    geo_data : shapely.geometry object
        Geo-spatial data with informations for location/region-shape. The
        geometry can be a polygon/multi-polygon for regions, a line fore
        transport objects or a point for objects such as transformer sources.

    Attributes
    ----------
    registry: :class:`EnergySystem <oemof.core.energy_system.EnergySystem>`
        The central registry keeping track of all :class:`Entities <Entity>`
        created. If this is `None`, :class:`Entity` instances are not
        kept track of. When you instantiate an :class:`EnergySystem
        <oemof.core.energy_system.EnergySystem>` it automatically becomes the
        entity registry, i.e. all entities created are added to its
        :attr:`entities <oemof.core.energy_system.EnergySystem.entities>`
        attribute on construction.
    """
    optimization_options = {}

    registry = None

    def __init__(self, **kwargs):
        # TODO: @Günni:
        # add default argument values to docstrings (if it's possible).
        self.uid = kwargs["uid"]
        self.inputs = kwargs.get("inputs", [])
        self.outputs = kwargs.get("outputs", [])
        for e_in in self.inputs:
            if self not in e_in.outputs:
                e_in.outputs.append(self)
        for e_out in self.outputs:
            if self not in e_out.inputs:
                e_out.inputs.append(self)
        self.geo_data = kwargs.get("geo_data", None)
        self.regions = []
        self.add_regions(kwargs.get('regions', []))
        if __class__.registry is not None:
            __class__.registry.entities.append(self)

        # TODO: @Gunni Yupp! Add docstring.
    def add_regions(self, regions):
        'Add regions to self.regions'
        self.regions.extend(regions)
        for region in regions:
            if self not in region.entities:
                region.entities.append(self)

    def __str__(self):
        # TODO: @Günni: Unused privat method. No Docstring.
        return "<{0} #{1}>".format(type(self).__name__, self.uid)



# -*- coding: utf-8 -*-
"""

@author: Simon Hilpert simon.hilpert@fh-flensburg.de
"""
import pandas as pd
import logging
from ..core.network.entities import Bus
from ..core.network.entities.components import sources as source
from ..core.network.entities.components import sinks as sink
from ..core.network.entities.components import transformers as transformer
from ..core.network.entities.components import transports as transport


def add_bus(row, **kwargs):
    r""" Adds bus object ot list of busses. The function is used
    by apply() method of pandas data frames.

    Parameters
    ----------
    row : row of pandas.DataFrame()
    busses : list
        list of existing busses
    busprices : pandas.DataFrame()
       data frame containting the values for the busprices. Column index is
       the uid of bus to create.
    """
    busprices = kwargs.get('busprices', None)
    busses = kwargs.get('busses', [])

    # check for prices
    if row['timeseries']:
        price = busprices[row['uid']]
    # if price is constant
    else:
        price = row['price']

    if row.get('sum_out_limit', False) == False or 'FALSE':
        sum_out_limit = float('+inf')
    else:
        sum_out_limit = float(row['sum_out_limit'])
    # set kwargs from row items
    kwargs = {}
    for k in row.keys():
        kwargs.update({k: row[k]})
    kwargs['price'] = price
    kwargs['sum_out_limit'] = sum_out_limit
    obj = Bus(**kwargs)
    busses.append(obj)


def add_source(row, **kwargs):
    r""" Adds source object ot list of source. The function is used
    by apply() method of pandas data frames.

    Parameters
    ----------
    row : row of pandas.DataFrame()
    busses : list
        list of existing busses
    sink : list
        list of sources where object are appended
    sourcevalues : pandas.DataFrame()
       data frame containting the values for the source column index is uid of
       source to create.
    """
    sourcevalues = kwargs.get('sourcevalues', None)
    busses = kwargs.get('busses', None)
    sources = kwargs.get('sources', None)

    cls = getattr(source, row['class'])

    # set kwargs from row
    kwargs = {}
    for k in row.keys():
        kwargs.update({k: row[k]})

    # set special kwargs (conversion to list etc. )
    kwargs['out_max'] = [row.get('out_max', None)]
    kwargs['outputs'] = [b for b in busses if b.uid == row['output']]
    kwargs['val'] = sourcevalues[row['uid']]

    obj = cls(**kwargs)
    sources.append(obj)


def add_sink(row, **kwargs):
    r""" Adds sink object ot list of transformers. The function is used
    by apply() method of pandas data frames.

    Parameters
    ----------
    row : row of pandas.DataFrame()
    busses : list
        list of existing busses
    sink : list
        list of sinks where object are appended
    sinkvalues : pandas.DataFrame()
       data frame containting the values for the sink column index is uid of
       sink
    """
    sinkvalues = kwargs.get('sinkvalues', None)
    busses = kwargs.get('busses', None)
    sinks = kwargs.get('sinks', [])

    cls = getattr(sink, row['class'])

    # set special kwargs (conversion to list etc. )
    inputs = [b for b in busses if b.uid == row['input']]

    # instantiate sink object
    obj = cls(uid=row['uid'],
              inputs=inputs,
              val=sinkvalues[row['uid']])
    sinks.append(obj)


def add_transformer(row, **kwargs):
    r""" Adds transformer objects ot list of transformers. The function is used
    by apply() method of pandas data frames.

    Parameters
    ----------
    row : row of pandas.DataFrame()
    busses : list
        list of existing busses
    transformers : list
        list of transformers where object are appended
    """
    if not row['skip']:
        busses = kwargs.get('busses', None)
        transformers = kwargs.get('transformers', [])
        # get transformer class

        cls = getattr(transformer, row['class'])

        # set kwargs from row
        kwargs = {}
        for k in row.keys():
            kwargs.update({k: row[k]})
        # set special kwargs (conversion to list etc. )
        kwargs['eta'] = [row['eta']]
        kwargs['out_max'] = [row['out_max']]
        kwargs['output_price'] = [row.get('output_price')]
        kwargs['outputs'] = [bus for bus in busses if bus.uid == row["output"]]
        kwargs['inputs'] = [b for b in busses if b.uid == row['input']]
        opex_var = kwargs['inputs'][0].price / kwargs['eta'][0]
        kwargs['opex_var'] = opex_var
        kwargs['out_min'] = [row['out_max']*row.get('out_min', 0)]

        # instantiate storage object from class
        obj = cls(**kwargs)
        transformers.append(obj)


def add_storage(row, **kwargs):
    r""" Adds storage objects ot list of storages. The function is used by
    apply() method of pandas data frames.

    Parameters
    ----------
    row : row of pandas.DataFrame()
    busses : list
        list of existing busses
    transformers : list
        list of transformers where object are appended
    """
    if not row['skip']:
        busses = kwargs.get('busses', None)
        transformers = kwargs.get('transformers', [])
        # get storage class
        cls = getattr(transformer, row['class'])
        # set kwargs from row
        kwargs = {}
        for k in row.keys():
            kwargs.update({k: row[k]})
        # set special kwargs (conversion to list etc. )
        kwargs['out_max'] = [row.get('out_max')]
        kwargs['in_max'] = [row.get('in_max')]
        kwargs['outputs'] = [bus for bus in busses if bus.uid == row["output"]]
        kwargs['inputs'] = [b for b in busses if b.uid == row['input']]

        # instantiate storage object from class
        obj = cls(**kwargs)
        transformers.append(obj)


def add_chp(row, **kwargs):
    r""" Adds chps objects ot list of transformers. The function is used by
    apply() method of pandas data frames.

    Parameters
    ----------
    row : row of pandas.DataFrame()
    busses : list
        list of existing busses
    transformers : list
        list of transformers where object are appended
    """
    if not row['skip']:
        busses = kwargs.get('busses', None)
        transformers = kwargs.get('transformers', None)
        cls = getattr(transformer, row['class'])

        # set kwargs from row
        kwargs = {}
        for k in row.keys():
            kwargs.update({k: row[k]})

        # special kwargs where transformation of type is necessary (e.g. list)
        kwargs['inputs'] = [b for b in busses if b.uid == row['input']]
        kwargs['outputs'] = [bus for bus in busses if
                             any([bus.uid == row["output_el"],
                                  bus.uid == row["output_th"]])]
        kwargs['out_max'] = [row['out_max_el'], row.get('out_max_th', None)]
        kwargs['out_min'] = [row['out_min']]
        kwargs['eta_min'] = [row['eta_el_min'], row.get('eta_th_min', None)]
        kwargs['output_price'] = [row.get('output_price_el'),
                                  row.get('output_price_th')]
        kwargs['eta'] = [row['eta_el'], row['eta_th']]
        # instantiate object with kwargs
        obj = cls(**kwargs)
        transformers.append(obj)


def add_transport(row, **kwargs):
    r""" Adds transpport objects ot list of transports. The function is used by
    apply() method of pandas data frames.

    """
    if not row['skip']:
        busses = kwargs.get('busses', None)
        transports = kwargs.get('transports', None)
        cls = getattr(transport, row['class'])

        # set kwargs from row
        kwargs = {}
        for k in row.keys():
            kwargs.update({k: row[k]})

        kwargs['eta'] = [row['eta']]
        kwargs['out_max'] = [row['out_max']]
        kwargs['in_max'] = [row.get('in_max')]
        kwargs['inputs'] = [b for b in busses if b.uid == row['input']]

        kwargs['outputs'] = [b for b in busses if b.uid == row['output']]

        # instantiate object with kwargs
        obj = cls(**kwargs)
        transports.append(obj)


def entities_from_csv(files, entities_dict=None):
    r""" Creates 'oemof-objects' from csv files by the use of pandas dataframes

    Parameters
    ----------
    files : dict
        dictionary containing the paths to files for object creation. Keys are:
        'transformers','busses','storages','chps', 'sources', 'sourcevalues',
        'sinks', 'sinkvalues', 'busprices'
    entities_dict : dict
        dictionary containing lists of oemof base class objects
    """
    if entities_dict is None:
        entities_dict = {'busses': [],
                         'transformers': [],
                         'sinks': [],
                         'sources': [],
                         'transports': []}
    # first create busses
    file = files.get('busses')
    if file is not None:
        bus_df = pd.read_csv(file)
        file = files.get('busprices')
        if file is not None:
            busprices = pd.read_csv(file)
        else:
            logging.info('No csv data for bus prices!')
            busprices = pd.DataFrame()
        bus_df.apply(add_bus, axis=1, busprices=busprices,
                     busses=entities_dict['busses'])
    # transformers
    file = files.get('transformers')
    if file is not None:
        df = pd.read_csv(file)
        df.apply(add_transformer, axis=1, busses=entities_dict['busses'],
                 transformers=entities_dict['transformers'])

    file = files.get('transports')
    if file is not None:
        df = pd.read_csv(file)
        df.apply(add_transport, axis=1, busses=entities_dict['busses'],
                 transports=entities_dict['transports'])

    file = files.get('storages')
    if file is not None:
        df = pd.read_csv(file)
        df.apply(add_storage, axis=1, busses=entities_dict['busses'],
                 transformers=entities_dict['transformers'])

    file = files.get('sources')
    if file is not None:
        df = pd.read_csv(file)
        values_csv = files.get('sourcevalues')
        if values_csv is not None:
            sourcevalues = pd.read_csv(values_csv)
        else:
            raise ValueError('No csv data found for source values!')
        df.apply(add_source, axis=1, busses=entities_dict['busses'],
                 sources=entities_dict['sources'], sourcevalues=sourcevalues)

    file = files.get('sinks')
    if file is not None:
        df = pd.read_csv(file)
        values_csv = files.get('sinkvalues')
        if values_csv is not None:
            sinkvalues = pd.read_csv(values_csv)
        else:
            raise ValueError('No csv data found for sink values!')
        df.apply(add_sink, axis=1, busses=entities_dict['busses'],
                 sinks=entities_dict['sinks'], sinkvalues=sinkvalues)

    file = files.get('chps', None)
    if file is not None:
        chp_df = pd.read_csv(file)
        chp_df.apply(add_chp, axis=1, busses=entities_dict['busses'],
                     transformers=entities_dict['transformers'])

    entities = sum([entities_dict[k] for k in entities_dict.keys()], [])
    entities_dict['entities'] = entities
    return(entities_dict)



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 17 11:08:15 2015

This is a collection of helper functions which work on there own an can be
used by various classes. If there are too many helper-functions, they will
be sorted in different modules.

All special import should be in try/except loops to avoid import errors.
"""

import logging
from datetime import date, timedelta
import os
import sys
import pandas as pd
import pickle
import time
import calendar
import pprint as pp
from oemof.demandlib import demand as dm


# get_polygon_from_shp_file
hlp_fkt = 'get_polygon_from_shp_file'
try:
    import fiona
    from shapely import geometry as shape
    from shapely.ops import cascaded_union
except:
    logging.info(
        'You will not be able to use the helper function: {0}'.format(hlp_fkt))
    logging.info('Install fiona, shapely to use it.')

# get_german_holidays
hlp_fkt = 'get_german_holidays'
try:
    from dateutil import easter
except:
    logging.info(
        'You will not be able to use the helper function: {0}'.format(hlp_fkt))
    logging.info('Install dateutil to use it.')

# fetch_admin_from_coord_osm
hlp_fkt = 'fetch_admin_from_coord_osm'
try:
    import urllib
    import xml.etree.ElementTree as ET
except:
    logging.info(
        'You will not be able to use the helper function: {0}'.format(hlp_fkt))
    logging.info('Install urllib, xml to use it.')

# fetch_admin_from_coord_google
hlp_fkt = 'fetch_admin_from_coord_google'
try:
    import geocoder
except:
    logging.info(
        'You will not be able to use the helper function: {0}'.format(hlp_fkt))
    logging.info('Install geocoder to use it.')

# Download a file from the internet
hlp_fkt = ' download_file'
try:
    from urllib.request import urlretrieve
except:
    logging.info(
        'You will not be able to use the helper function: {0}'.format(hlp_fkt))
    logging.info('Install urllib to use it.')


def get_polygon_from_shp_file(file):
    r"""A one-line summary that does not use variable names or the
    function name.

    Several sentences providing an extended description. Refer to
    variables using back-ticks, e.g. `var`.

    Parameters
    ----------
    var1 : array_like
        Array_like means all those objects -- lists, nested lists, etc. --
        that can be converted to an array.  We can also refer to
        variables like `var1`.
    var2 : int
        The type above can either refer to an actual Python type
        (e.g. ``int``), or describe the type of the variable in more
        detail, e.g. ``(N,) ndarray`` or ``array_like``.
    Long_variable_name : {'hi', 'ho'}, optional
        Choices in brackets, default first when optional.

    Returns
    -------
    type
        Explanation of anonymous return value of type ``type``.
    describe : type
        Explanation of return value named `describe`.
    out : type
        Explanation of `out`.

    Other Parameters
    ----------------
    only_seldom_used_keywords : type
        Explanation
    common_parameters_listed_above : type
        Explanation

    Raises
    ------
    BadException
        Because you shouldn't have done that.

    See Also
    --------
    otherfunc : relationship (optional)
    newfunc : Relationship (optional), which could be fairly long, in which
              case the line wraps here.
    thirdfunc, fourthfunc, fifthfunc

    Notes
    -----
    Notes about the implementation algorithm (if needed).

    This can have multiple paragraphs.

    You may include some math:

    .. math:: X(e^{j\omega } ) = x(n)e^{ - j\omega n}

    And even use a greek symbol like :math:`omega` inline.

    References
    ----------
    Cite the relevant literature, e.g. [1]_.  You may also cite these
    references in the notes section above.

    .. [1] O. McNoleg, "The integration of GIS, remote sensing,
       expert systems and adaptive co-kriging for environmental habitat
       modelling of the Highland Haggis using object-oriented, fuzzy-logic
       and neural-network techniques," Computers & Geosciences, vol. 22,
       pp. 585-588, 1996.

    Examples
    --------
    These are written in doctest format, and should illustrate how to
    use the function.

    >>> a=[1,2,3]
    >>> print([x + 3 for x in a])
    [4, 5, 6]
    >>> print("a\nb")
    a
    b

    """

    logging.debug('Loading polygon from file: {0}'.format(file))
    multi = shape.MultiPolygon(
        [shape.shape(pol['geometry']) for pol in fiona.open(file)])
    return cascaded_union(multi)


def get_german_holidays(year, place=[None, None, None], scope='legal'):
    r"""
    Returns German holiday dates. Use the abbreviations for the german
    federal states like 'BY' for Bayern.

    Parameters
    ----------
    year : int
        Year of which the holidays are wanted
    place : list of strings
        List of names: [country, state, city]
    scope : string
        Type of holidays. So far only legal is possible.

    Returns
    -------
    dictionary
        The keys are in the datetime format the values represent the names.

    Notes
    -----
    So far only the data of german holidays is available. Feel free to add
    other countries or to implement workalendar.
    https://github.com/novapost/workalendar

    Examples
    --------
    get_holidays(2015, place=['Germany', 'BE', 'Berlin'])
    """
    logging.debug('Feiertage für: {0}'.format(place))
    if place[0] not in ['Deutschland', 'deutschland', 'Germany', 'germany']:
        logging.error(
            'You are outside of Germany. The holidays will be incorrect.')
    # Determine Easter
    eastr = easter.easter(year)

    # Add national holidays
    adict = {
        date(year,  1,  1): 'Neujahr',
        date(year,  5,  1): 'Tag der Arbeit',
        date(year, 10,  3): 'Tag der Deutschen Einheit',
        date(year, 12, 25): '1. Weihnachtstag',
        date(year, 12, 26): '2. Weihnachtstag',
        eastr - timedelta(days=2): 'Karfreitag',
        eastr: 'Ostersonntag',
        eastr + timedelta(days=1): 'Ostermontag',
        eastr + timedelta(days=39): 'Christi Himmelfahrt',
        eastr + timedelta(days=50): 'Pfingstmontag',
        }

    # Add federal holidays
    if place[1].upper() in ('BW', 'BY', 'ST'):
        adict[date(year, 1, 6)] = 'Heilige Drei Könige'

    if place[1] in ('BW', 'BY', 'HE', 'NW', 'RP', 'SL'):
        adict[eastr + timedelta(days=60)] = 'Frohnleichnam'

    if place[1].upper() in ('BY', 'SL'):
        adict[date(year, 8, 15)] = 'Mariä Himmelfahrt'

    if place[1].upper() in ('BB', 'MV', 'SN', 'ST', 'TH'):
        adict[date(year, 10, 31)] = 'Reformationstag'

    if place[1].upper() in ('BW', 'BY', ):
        adict[date(year, 11, 1)] = 'Allerheiligen'

    return adict


def abbreviation_of_state(statename):
    r"""Get the abbreviation of a german state name or the other way round.

    Parameters
    ----------
    statename : string
        Fullname of the state or its abbreviation

    Returns
    -------
    string
        Fullname of the state or its abbreviation, depending on the input.
    """

    abbr_dict = {
        'Baden-Württemberg': 'BW',
        'Bayern': 'BY',
        'Berlin': 'BE',
        'Brandenburg': 'BB',
        'Bremen': 'HB',
        'Hamburg': 'HH',
        'Hessen': 'HE',
        'Mecklenburg-Vorpommern': 'MV',
        'Niedersachsen': 'NI',
        'Nordrhein-Westfalen': 'NW',
        'Rheinland-Pfalz': 'RP',
        'Saarland': 'SL',
        'Sachsen': 'SN',
        'Sachsen-Anhalt': 'ST',
        'Schleswig-Holstein': 'SH',
        'Thüringen': 'TH'}
    try:
        value = abbr_dict[statename]
    except:
        try:
            value = list(abbr_dict.keys())[list(abbr_dict.values()).index(
                statename)]
        except:
            value = None
    return value


def fetch_admin_from_coord_osm(coord):
    r"""Reverse geocoding using osm.

    Parameters
    ----------
    coord : array_like
        Geo-coordinates in the order (longitude, latitude)

    Returns
    -------
    list of strings
        Containing the name of the country and the name of the state if
        available

    See Also
    --------
    fetch_admin_from_coord_google

    Examples
    --------
    >>> fetch_admin_from_coord_osm((12.7, 51.8))
    ['Deutschland', 'ST']
    """

    def parse_result(res):
        root = ET.fromstring(res)
        address_parts = {}

        for a in root[1]:
            address_parts[a.tag] = a.text

        return address_parts

    lon = coord[0]
    lat = coord[1]

    query = "http://nominatim.openstreetmap.org/reverse?"
    query += "format=xml"
    query += "&lat={lat}".format(lat=lat)
    query += "&lon={lon}".format(lon=lon)
    query += "&zoom=18"
    query += "&addressdetails=1"

    conn = urllib.request.urlopen(query)
    rev_geocode = conn.read()
    address_parts = parse_result(rev_geocode)

    try:
        state = abbreviation_of_state(address_parts['state'])
    except KeyError:
        logging.error(
            "Didn't get the name of the state. " +
            "Maybe the coordinates ({0}) are outside of Germany.".format(
                str([lat, lon])))
        state = ''
    try:
        country = address_parts['country']
    except:
        country = None
    return [country, state]


def fetch_admin_from_coord_google(coord):
    r"""Reverse geocoding using google.

    Parameters
    ----------
    coord : array_like
        Geo-coordinates in the order (longitude, latitude)

    Returns
    -------
    list of strings
        Containing the name of the country and the name of the state if
        available

    See Also
    --------
    fetch_admin_from_coord_google

    Examples
    --------
    >>> fetch_admin_from_coord_osm((12.7, 51.8))
    ['Deutschland', 'ST']
    """

    new_coord = list((coord[1], coord[0]))
    g = geocoder.google(new_coord, method='reverse')
    return [g.country, g.state]


def remove_from_list(orig_list, remove_list):
    '''
    Removes the values inside the remove_list from the orig_list.
    '''
    for item in remove_list:
        if item in orig_list:
            try:
                orig_list.remove(item)
            except:
                logging.debug('Cannot remove %s from list %s' % (
                    item, orig_list))
    return orig_list


def cut_lists(list_a, list_b):
    '''
    Returns a list with the values of list_a AND list_b.
    '''
    return [x for x in list(list_a) if x in set(list_b)]


def unique_list(seq):
    '''
    Returns a unique list without preserving the order
    '''
    return list({}.fromkeys(seq).keys())


def time_logging(start, text, logging_level='debug'):
    '''
    Logs the time between the given start time and the actual time. A text
    and the debug level is variable.

    Uwe Krien (uwe.krien@rl-institut.de)

    Parameters
        start : start time : float
        text : text to describe the log : string

    Keyword arguments
        logging_level : logging_level [default='debug'] : string
    '''
    import time
    end_time = time.time() - start
    hours = int(end_time / 3600)
    minutes = int(end_time / 60 - hours * 60)
    seconds = int(end_time - hours * 3600 - minutes * 60)
    time_string = ' %0d:%02d:%02d hours' % (hours, minutes, seconds)
    log_str = text + time_string
    if logging_level == 'debug':
        logging.debug(log_str)
    elif logging_level == 'info':
        logging.info(log_str)


def mkdir_if_not_exist(path):
    '''Creates directory if not exist'''
    if not os.path.isdir(path):
        os.mkdir(path)


def unlist(val):
    '''Returns single value if a single value in a list is given'''
    if isinstance(val, list):
        if len(val) == 1:
            new_val = val[0]
    else:
        new_val = val
    return new_val


def dict2pickle(dic, filename=None, path=None):
    'Dumping a python dictionary into a pickle file.'
    if filename is None:
        filename = 'dict.pkl'
    if path is None:
        path = os.path.expanduser("~")
    logging.info('Writing parameters to {0}...'.format(
        os.path.join(path, filename)))
    outputfile = open(os.path.join(path, filename), 'wb')
    pickle.dump(dic, outputfile)
    outputfile.close()


def pickle2dict(filename=None, path=None):
    'Reading a python dictionary from a pickle file.'
    if filename is None:
        filename = 'dict.pkl'
    if path is None:
        path = os.path.expanduser("~")
    fullname = os.path.join(path, filename)
    logging.info('Reading parameters from {0} (created: {1})...'.format(
        fullname, time.ctime(os.path.getmtime(fullname))))
    inputfile = open(fullname, 'rb')
    dic = pickle.load(inputfile)
    inputfile.close()
    return dic


def call_demandlib(demand, method, year, **kwargs):
    '''
    Calls the demandlib and creates an object which includes the demand
    timeseries.

    Required Parameters
    -------------------
    demand :
    method : Method which is to be applied for the demand calculation
    '''

    df = create_basic_dataframe(year, **kwargs)
    demand.val = dm.electrical_demand(method,
                         dataframe=df,
                         ann_el_demand_per_sector=kwargs.get(
                         'ann_el_demand_per_sector'),
                         ann_el_demand_per_person=kwargs.get(
                         'ann_el_demand_per_person'),
                         household_structure=kwargs.get(
                         'household_structure'),
                         household_members_all=kwargs.get(
                         'household_members_all'),
                         population=kwargs.get(
                         'population'),
                         comm_ann_el_demand_state=kwargs.get(
                         'comm_ann_el_demand_state'),
                         comm_number_of_employees_state=kwargs.get(
                         'comm_number_of_employees_state'),
                         comm_number_of_employees_region=kwargs.get(
                         'comm_number_of_employees_region')).elec_demand

    return demand


def dict2textfile(dic, filename=None, path=None):
    'Writing a dictionary to textfile in a readable and clearly formatted way.'
    if filename is None:
        filename = 'dict2text.txt'
    if path is None:
        path = os.path.expanduser("~")
    logging.info('Writing formatted dictionary to {0}...'.format(
        os.path.join(path, filename)))
    f1 = open(os.path.join(path, filename), 'w+')
    pp.pprint(dic, f1)
    f1.close()


def download_file(filename, url):
    '''Copy a file from the given url to the given filename.
    '''
    if not os.path.isfile(filename):
        logging.info('Copying file from {0} to {1}'.format(
            url, filename))
        urlretrieve(url, filename)


def get_basic_path():
    if sys.platform == "win32":
        basicpath = os.path.join(os.environ['USERPROFILE'], '.oemof')
    else:  # if linux platform
        basicpath = os.path.join(os.environ['HOME'], '.oemof')
    if not os.path.isdir(basicpath):
        os.mkdir(basicpath)
    return basicpath


def extend_basic_path(subfolder):
    extended_path = os.path.join(get_basic_path(), subfolder)
    if not os.path.isdir(extended_path):
        os.mkdir(extended_path)
    return extended_path


def get_fullpath(path, filename):
    return os.path.join(path, filename)


def create_basic_dataframe(year, **kwargs):
    r"""Giving back a DataFrame containing weekdays and optionally holidays for the
    given year.

    Parameters
    ----------
    year: the year for which dataframe should be created

    Optional Parameters
    -------------------
    holidays: array with information for every hour of the year, if holiday or
        not (0: holiday, 1: no holiday)

    Returns
    -------
    pandas.DataFrame : DataFrame with a time index

    Notes
    -----
    Using Pandas > 0.16

    """
    if calendar.isleap(year):
        hoy = 8784
    else:
        hoy = 8760

    time_df = pd.DataFrame(
        index=pd.date_range(pd.datetime(year, 1, 1, 0), periods=hoy,
                            freq='H'),
        columns=['weekday', 'hour', 'date'])

    # Add a column 'hour of the day to the DataFrame
    time_df['hour'] = time_df.index.hour + 1
    time_df['weekday'] = time_df.index.weekday + 1
    time_df['date'] = time_df.index.date

    # Set weekday to Holiday (0) for all holidays
    if kwargs.get('holidays'):
        time_df['weekday'].mask(pd.to_datetime(time_df['date']).isin(
            pd.to_datetime(list(kwargs['holidays'].keys()))), 0, True)

#    holidays = helpers.get_german_holidays(year, place)

    df = time_df

    return df



# -*- coding: utf-8 -*-
"""
Module contains variable definitions and constraint to bound variables (e.g.
for investement).

@author: Simon Hilpert (simon.hilpert@fh-flensburg.de)
"""

import pyomo.environ as po
import numpy as np
import logging


def add_binary(model, block, relaxed=False):
    """ Creates all status variables (binary) for `block.objs`

    Status variable indicates if a unit is turned on or off.
    E.g. if a transformer is switched on/off -> y=1/0

    Parameters
    ----------
    model : pyomo.ConcreteModel()
        A pyomo-object to be solved containing all Variables, Constraints, Data
        Variables are added as attributes to the `model`
    objs : array_like (list)
        all components for which the status variable is created
    relaxed : boolean
       If True "binary" variables will be created as continuous variables with
       bounds of 0 and 1.

    """
    # check
    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                          which the status variable should be created.")
    # add binary variables to model
    if not relaxed:
        block.y = po.Var(block.indexset, within=po.Binary)
    if relaxed:
        block.y = po.Var(block.indexset, within=po.NonNegativeReals,
                         bounds=(0, 1))


def add_continuous(model, edges):
    """ Adds all variables corresponding to the edges of the bi-partite
    graph for all timesteps.

    The function uses the pyomo class `Var()` to create the optimization
    variables. As index-sets the provided edges of the graph
    (in general all edges) and the defined timesteps are used.
    The following variables are created: Variables for all the edges
    (If specific components such as disptach sources and storages exist.)

    Parameters
    ----------
    model : OptimizationModel() instance
        A object to be solved containing all Variables, Constraints, Data
        Variables are added as attributes to the `model`
    edges : array_like (list)
        `edges` will be a list containing tuples representing the directed
        edges of the graph by using unique ids of components and buses.
        e.g. [('coal', 'pp_coal'), ('pp_coal', 'b_el'),...]
    """
    # variable for all edges
    model.w = po.Var(edges, model.timesteps, within=po.NonNegativeReals)


def set_bounds(model, block, side='output'):
    """ Sets bounds for variables that represent the weight
    of the edges of the graph if investment models are calculated.

    For investment  models upper and lower bounds will be modeled via
    additional constraints. The mathematical description for the
    constraint is as follows:

    If side is `output`

    .. math:: w_{e, o_{e,1}}(t) \\leq \\overline{W}_{e, o_{e,1}} \\qquad \
    \\forall e, \\forall t

    With investment:

    .. math::  w_{e, o_{e,1}}(t) \\leq \\overline{W}_{e, o_{e,1}} + \
    \\overline{w}^{add}_{o_e}, \\qquad \\forall e, \\forall t

    If side is `input`:

    .. math:: w_{i_{e}, e}(t) \\leq \\overline{W}_{i_{e}, e} \\qquad \
    \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing
    the set of unique ids for all entities grouped inside the
    attribute `block.objs`.


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data
        Constraints are added as attributes to the `model`
    block : SimpleBlock()
    side : string
       Side of component for which the bounds are set ('input' or 'output')

    """
    logging.debug("Set bounds for {0}.".format(block))
    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                          which bounds should be set.")
    if block.uids is None:
        block.uids = [e.uids for e in block.objs]

    # set variable bounds (out_max = in_max * efficiency):
    # m.in_max = {'pp_coal': 51794.8717948718, ... }
    # m.out_max = {'pp_coal': 20200, ... }
    ub_in = {}
    ub_out = {}
    out_max = {}
    exist_ub_out = False
    for e in block.objs:
        if side == 'output':
            output_uids = [o.uid for o in e.outputs[:]]
            # ** Time depended bound
            if e.ub_out:
                ub_out[e.uid] = dict(zip(output_uids, e.ub_out))
                out_max[e.uid] = dict(zip(output_uids, e.out_max))
                if e.out_max < np.array(e.ub_out).max():
                    logging.error('The maximal value of ub_out should not be' +
                                  ' greater than out_max ({}).'.format(e.uid))
                exist_ub_out = True
            # ** Constant bound
            else:
                # If ub_out (time depended bound9 does not exist, create a list
                # with the same value. Use out_max for the upper bound of the
                # variable.
                ub_out[e.uid] = dict(zip(
                    output_uids,
                    [[x] * len(model.timesteps) for x in e.out_max]))
                out_max[e.uid] = dict(zip(output_uids, e.out_max))
        if side == 'input' and e.in_max is not None:
            input_uids = [i.uid for i in e.inputs[:]]
            ub_in[e.uid] = dict(zip(
                input_uids,
                [[x] * len(model.timesteps) for x in e.in_max]))

    # *** No investment - set upper bound to maximal output***
    if not block.optimization_options.get('investment', False):
        # edges for simple transformers ([('coal', 'pp_coal'),...])
        ee = model.edges(block.objs)
        for (e1, e2) in ee:
            for t in model.timesteps:
                # transformer output <= model.out_max
                if e1 in block.uids and side == 'output':
                    model.w[e1, e2, t].setub(ub_out[e1][e2][t])
                # transformer input <= model.in_max
                if e2 in ub_in and side == 'input':
                    try:
                        model.w[e1, e2, t].setub(ub_in[e2][e1][t])
                    except:
                        logging.warning("No upper bound for input (%s,%s)",
                                        e1, e2)
                        pass

    # *** Investment - set upper bound to overall output and add constraint***
    else:
        if side == 'output':
            # set maximum of addiational storage capacity
            add_out_limit = {obj.uid: obj.add_out_limit for obj in block.objs}
            # loop over all uids (storages) set the upper bound
            for e in block.uids:
                block.add_out[e].setub(add_out_limit[e])

            # constraint for additional capacity
            def add_output_rule(block, e, t):
                lhs = model.w[e, model.O[e][0], t]
                rhs = ub_out[e][model.O[e][0]][t] + block.add_out[e]
                return(lhs <= rhs)

            # constraint for additional capacity
            def add_output_rule_time_depended_bound(block, e, t):
                lhs = model.w[e, model.O[e][0], t]
                rhs = ub_out[e][model.O[e][0]][t] * (
                    1 + block.add_out[e] / out_max[e][model.O[e][0]])
                return(lhs <= rhs)

            if exist_ub_out:
                block.output_bound = po.Constraint(
                        block.indexset,
                        rule=add_output_rule_time_depended_bound)
            else:
                block.output_bound = po.Constraint(
                        block.indexset, rule=add_output_rule)

        # TODO: Implement upper bound constraint for investment models
        elif side == 'input':
            error_mesg = "Setting upper bounds on inputs of components is "
            error_mesg += "not possible for investment models"
            for e in block.objs:
                if e.in_max is not None:
                    for in_max in e.in_max:
                        if in_max is not None and in_max != float("inf"):
                            raise ValueError(error_mesg)


def set_storage_cap_bounds(model, block):
    """ Alters/sets upper and lower bounds for variables that represent the
    absolut state of charge e.g. filling level of a storage component.

    For investment  models upper and lower bounds will be modeled via
    additional constraints. The mathematical description for the
    constraint is as follows:

     .. math:: \\underline{L}_e \\leq l_e(t) \\leq \\overline{L}_e, \
    \\qquad \\forall e, \\forall t

    If investment:

    .. math:: l_e(t)  \\leq \\overline{L}_e + \\overline{l}^{add}_e, \
    \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing
    the set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally :math:`\mathcal{E} \subset \mathcal{E}_{S}`.

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data
        Bounds are altered at model attributes (variables) of `model`
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    """

    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                         which bounds should be set.")
    if block.uids is None:
        block.uids = [e.uids for e in block.objs]

    # extract values for storages m.cap_max = {'storge': 120.5, ... }
    cap_max = {obj.uid: obj.cap_max for obj in block.objs}
    cap_min = {obj.uid: obj.cap_min for obj in block.objs}

    if not block.optimization_options.get('investment', False):
        # loop over all uids (storages) and timesteps to set the upper bound
        for e in block.uids:
            for t in model.timesteps:
                block.cap[e, t].setub(cap_max[e])
                block.cap[e, t].setlb(cap_min[e])
    else:
        # set maximum of additional storage capacity
        add_cap_limit = {obj.uid: obj.add_cap_limit for obj in block.objs}
        # loop over all uids (storages) set the upper bound
        for e in block.uids:
            block.add_cap[e].setub(add_cap_limit[e])

        # constraint for additional capacity in investment models
        def add_cap_rule(block, e, t):
            lhs = block.cap[e, t]
            rhs = cap_max[e] + block.add_cap[e]
            return(lhs <= rhs)
        block.cap_bound = po.Constraint(block.indexset,
                                        rule=add_cap_rule)


def set_outages(model, block, outagetype='period', side='output'):
    """ Fixes component input/output to zeros for modeling outages.


    Parameters
    ----------
    model :OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    outagetype : string
        Type to model outages of component if outages is scalar.
       'period' yield one timeblock where component is off,
       while 'random_days' will sample random days over the timehorizon
       where component is off
    side : string
       Side of component to fix to zero: 'output', 'input'.

    """
    outages = {obj.uid: obj.outages for obj in block.objs}
    timesteps = {}
    for e in outages.keys():
        if isinstance(outages[e], (float, np.float)) \
                and outages[e] <= 1 and outages[e] >= 0:
            time_off = int(len(model.timesteps) * outages[e])
            if outagetype == 'period':
                start = np.random.randint(0, len(model.timesteps)-time_off+1)
                end = start + time_off
                timesteps[e] = [t for t in range(start, end)]
            if outagetype == 'random_days':
                timesteps[e] = np.random.randint(0, len(model.timesteps))
        elif len(outages[e]) >= 1:
            timesteps[e] = outages[e]
        else:
            timesteps[e] = []

    if side == 'input' and timesteps[e]:
        for e in block.uids:
            for t in timesteps[e]:
                model.w[model.I[e][0], e, t] = 0
                model.w[model.I[e][0], e, t].fix()
    if side == 'output' and timesteps[e]:
        for e in block.uids:
            for t in timesteps[e]:
                model.w[e, model.O[e][0], t] = 0
                model.w[e, model.O[e][0], t].fix()
    else:
        pass


def set_fixed_sink_value(model, block):
    """ Setting a value und fixes the variable of input.

    The mathematical formulation is as follows:

    .. math:: w_{i_e, e}(t) = V_e(t), \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing
    the set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally :math:`\mathcal{E} \subset \mathcal{E}_{I}`.

    Parameters
    ----------
    model :OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data
        Attributes are altered of the `model`
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    """

    val = {obj.uid: obj.val for obj in block.objs}
    ee = model.edges(block.objs)
    for (e1, e2) in ee:
        for t in model.timesteps:
            # set variable value
            model.w[(e1, e2), t] = val[e2][t]
            # fix variable value for optimization problem
            model.w[(e1, e2), t].fix()



#! /usr/bin/env python

"""TODO: Maybe add a docstring containing a long description for oemof?

  This would double as something we could put int the `long_description`
  parameter for `setup` and it would squelch some complaints pylint has on
  `setup.py`.

"""

from setuptools import find_packages, setup

import oemof

setup(name='oemof',
      version=oemof.__version__,
      author='oemof developing group',
      author_email='oemof@rl-institut.de',
      description='The open energy modelling framework',
      namespace_package = ['oemof'],
      packages=find_packages(),
      package_dir={'oemof': 'oemof'},
      install_requires=['numpy >= 1.7.0',
                        'pandas >= 0.17.0',
                        'pyomo >= 4.2.0']
     )



# -*- coding: utf-8 -*-

'''Holiday information for Germany.'''

import logging
from datetime import date, timedelta
from dateutil import easter
import urllib
import xml.etree.ElementTree as ET


def get_german_holidays(year, place=[None, None, None], scope='legal', _=str):
    """
    Returns German holiday dates. Use the s for the german
    federal states like 'Germany/BY ' for Bayern.
    :
    Example: get_holidays(2015, place=['Germany', 'BE', 'Berlin'])
    """
    logging.debug('Feiertage für: {0}'.format(place))
    if place[0] not in ['Deutschland', 'deutschland', 'Germany', 'germany']:
        logging.error(
            'You are outside of Germany. The holidays will be incorrect.')
    # Determine Easter
    eastr = easter.easter(year)

    # Add national holidays
    adict = {
        date(year,  1,  1): 'Neujahr',
        date(year,  5,  1): 'Tag der Arbeit',
        date(year, 10,  3): 'Tag der Deutschen Einheit',
        date(year, 12, 25): '1. Weihnachtstag',
        date(year, 12, 26): '2. Weihnachtstag',
        eastr - timedelta(days=2): 'Karfreitag',
        eastr: 'Ostersonntag',
        eastr + timedelta(days=1): 'Ostermontag',
        eastr + timedelta(days=39): 'Christi Himmelfahrt',
        eastr + timedelta(days=50): 'Pfingstmontag',
        }

    # Add federal holidays
    if place[1].upper() in ('BW', 'BY', 'ST'):
        adict[date(year, 1, 6)] = 'Heilige Drei Könige'

    if place[1] in ('BW', 'BY', 'HE', 'NW', 'RP', 'SL'):
        adict[eastr + timedelta(days=60)] = 'Frohnleichnam'

    if place[1].upper() in ('BY', 'SL'):
        adict[date(year, 8, 15)] = 'Mariä Himmelfahrt'

    if place[1].upper() in ('BB', 'MV', 'SN', 'ST', 'TH'):
        adict[date(year, 10, 31)] = 'Reformationstag'

    if place[1].upper() in ('BW', 'BY', ):
        adict[date(year, 11, 1)] = 'Allerheiligen'

    return adict


def fetch_admin_from_latlon(lat=None, lon=None, coord=None):
    """
    Receive reverse geocoded information from lat/lon point
    :rtype : dict
    :param lat: latitude
    :param lon: longitude
    :param zoom: detail level of information
    :return: list: [country, state]
    """
    def parse_result(res):
        root = ET.fromstring(res)
        address_parts = {}

        for a in root[1]:
            address_parts[a.tag] = a.text

        return address_parts

    if lat is None or lon is None:
        lon = coord[0]
        lat = coord[1]

    query = "http://nominatim.openstreetmap.org/reverse?"
    query += "format=xml"
    query += "&lat={lat}".format(lat=lat)
    query += "&lon={lon}".format(lon=lon)
    query += "&zoom=18"
    query += "&addressdetails=1"

    conn = urllib.request.urlopen(query)
    rev_geocode = conn.read()
    address_parts = parse_result(rev_geocode)

    try:
        state = get_abbreviation_of_state(address_parts['state'])
    except KeyError:
        logging.error(
            "Didn't get the name of the state. " +
            "Maybe the coordinates ({0}) are outside of Germany.".format(
                str([lat, lon])))
        state = ''

    try:
        country = address_parts['country']
    except:
        country = ''

    return [country, state]


def get_abbreviation_of_state(statename):
    abbr_dict = {
        'Baden-Württemberg': 'BW',
        'Bayern': 'BY',
        'Berlin': 'BE',
        'Brandenburg': 'BB',
        'Bremen': 'HB',
        'Hamburg': 'HH',
        'Hessen': 'HE',
        'Mecklenburg-Vorpommern': 'MV',
        'Niedersachsen': 'NI',
        'Nordrhein-Westfalen': 'NW',
        'Rheinland-Pfalz': 'RP',
        'Saarland': 'SL',
        'Sachsen': 'SN',
        'Sachsen-Anhalt': 'ST',
        'Schleswig-Holstein': 'SH',
        'Thüringen': 'TH'}
    return abbr_dict[statename]


if __name__ == "__main__":
    from pprint import pprint
    lat = 48.
    lon = 11.5
    pprint(get_german_holidays(2015, place=fetch_admin_from_latlon(lat, lon)))



# -*- coding: utf-8 -*-
"""
This module contains predefined objectives that can be used to model
energy systems.

@author: Simon Hilpert (simon.hilpert@fh-flensburg.de)
"""
try:
    import objective_expressions as objexpr
except:
    from . import objective_expressions as objexpr

import pyomo.environ as po
import oemof.solph as solph

from ..core.network.entities import Bus, ExcessSlack, ShortageSlack
from ..core.network.entities.components import transformers as transformer
from ..core.network.entities.components import sources as source


def minimize_cost(self, cost_objects=None, revenue_objects=None):
    """ Builds objective function that minimises the total costs.

    Costs included are:
                        opex_var,
                        opex_fix,
                        curtailment_costs (dispatch sources),
                        annualised capex (investment components)

    Parameters
    ----------
    self : pyomo model instance
    cost_blocks : array like
       list containing classes of objects that are included in
               cost terms of objective function
    revenue_blocks : array like
       list containing classes of objects that are included in revenue
               terms of objective function
    """
    expr = 0
    c_blocks = cost_objects
    r_blocks = revenue_objects

    if cost_objects is None:
        c_blocks = [str(transformer.Simple),
                    str(transformer.PostHeating),
                    str(transformer.VariableEfficiencyCHP),
                    str(transformer.SimpleExtractionCHP),
                    str(transformer.Storage),
                    str(transformer.CHP),
                    str(source.FixedSource),
                    str(source.Commodity)]
    if revenue_objects is None:
        r_blocks = []

    blocks = [block for block in self.block_data_objects(active=True)
              if not isinstance(block,
                                solph.optimization_model.OptimizationModel)]

    for block in blocks:
        if block.name in c_blocks:
            if block.name == str(transformer.Storage):
                ref = 'capacity'
                expr += objexpr.add_opex_var(self,
                                             getattr(self,
                                                     str(transformer.Storage)),
                                             ref='input')
            else:
                ref = 'output'
            # variable costs

            expr += objexpr.add_opex_var(self, block, ref='output')
            # fix costs
            if block != str(source.Commodity):
                expr += objexpr.add_opex_fix(self, block, ref=ref)
            # investment costs
            if block.optimization_options.get('investment', False):
                expr += objexpr.add_capex(self, block, ref=ref)
            if hasattr(block, 'z_start'):
                expr += objexpr.add_startup_costs(self, block)

            # revenues
        if block.name in r_blocks:
            expr += objexpr.add_revenues(self, block, ref='output')

    # costs for dispatchable sources
    if hasattr(self, str(source.DispatchSource)):
        expr += \
            objexpr.add_curtailment_costs(self,
                                          getattr(self,
                                                  str(source.DispatchSource)))

    # artificial costs for excess or shortage
    if hasattr(self, str(ExcessSlack)):
        expr += objexpr.add_excess_slack_costs(self,
                                               getattr(self, str(ExcessSlack)))
    if hasattr(self, str(ShortageSlack)):
        expr += objexpr.add_shortage_slack_costs(self,
                                                 getattr(self, str(ShortageSlack)))


    self.objective = po.Objective(expr=expr)

